/* Name: powerSwitch.c
 * Project: PowerSwitch based on AVR USB driver
 * Author: Christian Starkjohann
 * Creation Date: 2005-01-16
 * Tabsize: 4
 * Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id: powerSwitch.c 472 2008-01-21 18:21:59Z cs $
 */

/*
General Description:
This program controls the PowerSwitch USB device from the command line.
It must be linked with libusb, a library for accessing the USB bus from
Linux, FreeBSD, Mac OS X and other Unix operating systems. Libusb can be
obtained from http://libusb.sourceforge.net/.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <usb.h>    /* this is libusb, see http://libusb.sourceforge.net/ */
#include <syslog.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <arpa/inet.h>

#include <stdint.h>

#include <linux/udp.h>
#include <linux/ip.h>
#include <string.h>
#include <syslog.h>
#include "irUtil.h"

#define USBDEV_SHARED_VENDOR    0x16C0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   0x05DC  /* Obdev's free shared PID */
/* Use obdev's generic shared VID/PID pair and follow the rules outlined
 * in firmware/usbdrv/USBID-License.txt.
 */

#define PSCMD_ECHO  0
#define PSCMD_RESET 1
#define PSCMD_SEND  2
#define PSCMD_SENDCMD  3
#define PSCMD_WARMUP 4
/* These are the vendor specific SETUP commands implemented by our USB device */

static void usage(char *name)
{
  syslog(LOG_USER|LOG_ERR, "usage:\n");
  syslog(LOG_USER|LOG_ERR, "  %s status\n", name);
  syslog(LOG_USER|LOG_ERR, "  %s get <port> [<duration>]\n", name);
  syslog(LOG_USER|LOG_ERR, "  %s set <port> [<duration>]\n", name);
}


static int  usbGetStringAscii(usb_dev_handle *dev, int index, int langid, char *buf, int buflen)
{
  char    buffer[256];
  int     rval, i;

  if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + index, langid, buffer, sizeof(buffer), 1000)) < 0)
    return rval;
  if(buffer[1] != USB_DT_STRING)
    return 0;
  if((unsigned char)buffer[0] < rval)
    rval = (unsigned char)buffer[0];
  rval /= 2;
  /* lossy conversion to ISO Latin1 */
  for(i=1; i<rval; i++) {
    if(i > buflen)  /* destination buffer overflow */
      break;
    buf[i-1] = buffer[2 * i];
    if(buffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
      buf[i-1] = '?';
  }
  buf[i-1] = 0;
  return i-1;
}


/* PowerSwitch uses the free shared default VID/PID. If you want to see an
 * example device lookup where an individually reserved PID is used, see our
 * RemoteSensor reference implementation.
 */

#define USB_ERROR_NOTFOUND  1
#define USB_ERROR_ACCESS    2
#define USB_ERROR_IO        3

static int usbOpenDevice(usb_dev_handle **device, int vendor, char *vendorName, int product, char *productName)
{
  struct usb_bus      *bus;
  struct usb_device   *dev;
  usb_dev_handle      *handle = NULL;
  int                 errorCode = USB_ERROR_NOTFOUND;
  static int          didUsbInit = 0;

  if(!didUsbInit) {
    didUsbInit = 1;
    usb_init();
  }
  usb_find_busses();
  usb_find_devices();
  for(bus=usb_get_busses(); bus; bus=bus->next) {
    for(dev=bus->devices; dev; dev=dev->next) {
      if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product) {
        char    string[256];
        int     len;
        handle = usb_open(dev); /* we need to open the device in order to query strings */
        if(!handle) {
          errorCode = USB_ERROR_ACCESS;
          syslog(LOG_USER|LOG_ERR, "Warning: cannot open USB device: %s\n", usb_strerror());
          continue;
        }
        if(vendorName == NULL && productName == NULL) { /* name does not matter */
          break;
        }
        /* now check whether the names match: */
        len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
        if(len < 0) {
          errorCode = USB_ERROR_IO;
          syslog(LOG_USER|LOG_ERR, "Warning: cannot query manufacturer for device: %s\n", usb_strerror());
        } else {
          errorCode = USB_ERROR_NOTFOUND;
          /* syslog(LOG_USER|LOG_ERR, "seen device from vendor ->%s<-\n", string); */
          if(strcmp(string, vendorName) == 0) {
            len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
            if(len < 0) {
              errorCode = USB_ERROR_IO;
              syslog(LOG_USER|LOG_ERR, "Warning: cannot query product for device: %s\n", usb_strerror());
            } else {
              errorCode = USB_ERROR_NOTFOUND;
              /* syslog(LOG_USER|LOG_ERR, "seen product ->%s<-\n", string); */
              if(strcmp(string, productName) == 0)
                break;
            }
          }
        }
        usb_close(handle);
        handle = NULL;
      }
    }
    if(handle)
      break;
  }
  if(handle != NULL) {
    errorCode = 0;
    *device = handle;
  }
  return errorCode;
}

static usb_dev_handle* g_handle;
void ir_init(void)
{
  usb_init();
  if(usbOpenDevice(&(g_handle), USBDEV_SHARED_VENDOR, "kiesbauer", USBDEV_SHARED_PRODUCT, "irUtil") != 0) {
    syslog(LOG_USER|LOG_ERR, "Could not find USB device \"test\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
  }
}

void run_cmd_send(usb_dev_handle* handle)
{
  char buffer[8];
  int ret = -1;
  ret = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_SEND, 0, 0, buffer, sizeof(buffer), 5000);
  if(ret < 0) {
    syslog(LOG_USER | LOG_WARNING, "Failed to send usb_controle_msg: rv: (%d): %s\n", ret, strerror(ret));
  }
}

void ir_sendSwitchCommand(bool on)
{
  if(on){
    run_cmd_send(g_handle);
    run_cmd_send(g_handle);
  } else {
    run_cmd_send(g_handle);
    run_cmd_send(g_handle);
    usleep(3 * 1000 * 1000);
    run_cmd_send(g_handle);
    run_cmd_send(g_handle);
    usleep(3 * 1000 * 1000);
    run_cmd_send(g_handle);
    run_cmd_send(g_handle);
  }
}


int __main(int argc, char **argv)
{
  //daemon(0,0);
  usb_dev_handle      *handle = NULL;
  unsigned char       buffer[8];
  int                 nBytes;

  openlog(argv[0], LOG_PID, LOG_USER);
  if(argc < 2) {
    usage(argv[0]);
    exit(1);
  }
  usb_init();
  if(usbOpenDevice(&handle, USBDEV_SHARED_VENDOR, "kiesbauer", USBDEV_SHARED_PRODUCT, "irUtil") != 0) {
    syslog(LOG_USER|LOG_ERR, "Could not find USB device \"test\" with vid=0x%x pid=0x%x\n", USBDEV_SHARED_VENDOR, USBDEV_SHARED_PRODUCT);
    exit(1);
  }
  /* We have searched all devices on all busses for our USB device above. Now
   * try to open it and perform the vendor specific control operations for the
   * function requested by the user.
   */
  if(strcmp(argv[1], "test") == 0) {
    int i, v, r;
    /* The test consists of writing 1000 random numbers to the device and checking
     * the echo. This should discover systematic bit errors (e.g. in bit stuffing).
     */
    for(i=0; i<1000; i++) {
      v = rand() & 0xffff;
      nBytes = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_ECHO, v, 0, (char *)buffer, sizeof(buffer), 5000);
      if(nBytes < 2) {
        if(nBytes < 0)
          syslog(LOG_USER|LOG_ERR, "USB error: %s\n", usb_strerror());
        syslog(LOG_USER|LOG_ERR, "only %d bytes received in iteration %d\n", nBytes, i);
        syslog(LOG_USER|LOG_ERR, "value sent = 0x%x\n", v);
        exit(1);
      }
      r = buffer[0] | (buffer[1] << 8);
      if(r != v) {
        syslog(LOG_USER|LOG_ERR, "data error: received 0x%x instead of 0x%x in iteration %d\n", r, v, i);
        exit(1);
      }
    }
    printf("test succeeded\n");
  } else if(strcmp(argv[1], "start") == 0) {
    run_cmd_send(handle);
    run_cmd_send(handle);
    usleep(3 * 1000 * 1000);
    run_cmd_send(handle);
    run_cmd_send(handle);
    usleep(3 * 1000 * 1000);
    run_cmd_send(handle);
    run_cmd_send(handle);
  } else if(strcmp(argv[1], "stop") == 0) {
    run_cmd_send(handle);
    run_cmd_send(handle);
  } else if(strcmp(argv[1], "reset") == 0) {
    char buffer[5];
    usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_RESET, 0, 0, buffer, sizeof(buffer), 5000);
  } else if(strcmp(argv[1], "cmd") == 0 && argc > 1) {
    int cmd = atoi(argv[1]);
    int len=0;
    len = usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 3, cmd, 0, (char*)buffer, sizeof(buffer), 5000);

    if(len == 2 && buffer[0] == 0xfe && buffer[1] == cmd)
      printf("Command was sent successfully!\n");
    else
      printf("Failed to send command! len: %d v1: %X v2: %X v3: %X cmd: %d\n", len, buffer[0], buffer[1], buffer[2], cmd);
  } else if(strcmp(argv[1], "warmup") == 0) {
    char buffer[5];
    usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PSCMD_WARMUP, 0, 0, buffer, sizeof(buffer), 5000);
  } else if(strcmp(argv[1], "pwm") == 0) {
    char buffer[5];
    usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 5, atoi(argv[2]), 0, buffer, sizeof(buffer), 5000);
    printf("written %d\n", (unsigned char)buffer[0]);
  } else if(strcmp(argv[1], "daemon") == 0) {

    int appletvcontrolFd = open("/tmp/appletvcontrol", O_RDONLY);
    if(appletvcontrolFd < 0) {
      perror("Failed to open");
    }
    int statusFd = open("/tmp/irStatus", O_RDWR | O_CREAT);

    while(1) {
      uint8_t fromAppleTv = 0;
      if(read(appletvcontrolFd, &fromAppleTv, 1) < 0) {
        printf("Read a byte\n");
        write(statusFd, &fromAppleTv, 1);
        lseek(statusFd, 0, SEEK_SET);
      } else {
        perror("Failed to read");
      }
    }

  }
  usb_close(handle);
  return 0;
}
